/*------------------------------------------------------------------------------
* convbin.c : convert receiver binary log file to rinex obs/nav, sbas messages
*
*          Copyright (C) 2007-2009 by T.TAKASU, All rights reserved.
*
* options : -DWIN32 use windows file path separator
*
* version : $Revision: 1.1 $ $Date: 2008/07/17 22:13:04 $
* history : 2008/06/22 1.0 new
*           2009/06/17 1.1 support glonass
*-----------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include "rtklib.h"

static const char rcsid[]="$Id: convbin.c,v 1.1 2008/07/17 22:13:04 ttaka Exp $";

#define PRGNAME "CONVBIN"

#ifdef WIN32
#define PATHSEP     '\\'            /* file path separator */
#else
#define PATHSEP     '/'             /* file path separator */
#endif

/* help text -----------------------------------------------------------------*/
static const char *help[]={
"",
" Synopsys",
"",
" convbin [-ts y/m/d h:m:s] [-te y/m/d h:m:s] [-ti tint] [-r rcv] [-f ffreq]",
"         [-od] [-os] [-d dir] [-o ofile] [-n nfile] [-g gfile] [-s sfile] file",
"",
" Description",
"",
" Convert receiver raw data log file to RINEX OBS/NAV and SBAS message file.",
" SBAS message file complies to RTKLIB SBAS messsage format. Support the",
" following message formats.",
"",
" RTCM 2               : Type 1, 3, 9, 14, 16, 17, 18, 19, 22",
" RTCM 3               : Type 1002, 1004, 1005, 1006, 1010, 1012, 1019, 1020",
" NovAtel OEMV/4       : RANGECMPB, RANGEB, RAWEPHEMB, IONUTCB, RAWWASSFRAMEB",
" NovAtel OEM3         : RGEB, REGD, REPB, FRMB, IONB, UTCB",
" u-blox LEA-4T/LEA-5T : RXM-RAW, RXM-SFRB",
" NovAtel Superstar II : ID#20, ID#21, ID#22, ID#23, ID#67",
" Hemisphere           : BIN76, BIN80, BIN94, BIN95, BIN96",
" SkyTraq S1315F       : msg0xDD, msg0xE0, msg0xDC",
"",
" Options [default]",
"",
"     file      input receiver binary log file",
"     -ts y/m/d h:m:s  start time [all]",
"     -te y/m/d h:m:s  end time [all]",
"     -ti tint  observation data interval (s) [all]",
"     -r format log format type",
"               rtcm2= RTCM 2",
"               rtcm3= RTCM 3",
"               nov  = NovAtel OEMV/4",
"               oem3 = NovAtel OEM3",
"               ubx  = ublox LEA-4T/LEA-5T",
"               ss2  = NovAtel Superstar II",
"               hemis= Hemisphere Eclipse/Crescent",
"               stq  = SkyTraq S1315F",
"     -f freq   number of frequencies [2]",
"     -od       include doppler frequency [off]",
"     -os       include snr [off]",
"     -d dir    output directory [same as input file]",
"     -o ofile  output RINEX OBS file    [<file>.obs]",
"     -n nfile  output RINEX NAV file    [<file>.nav]",
"     -g gfile  output RINEX GNAV file   [<file>.gnav]",
"     -s sfile  output SBAS message file [<file>.sbs]",
"",
" If receiver type is not specified, type is recognized by the input",
" file extention as follows.",
"     *.rtcm2  RTCM 2",
"     *.rtcm3  RTCM 3",
"     *.gps    NovAtel OEMV/4",
"     *.ubx    u-blox LEA-4T/LEA-5T",
"     *.log    NovAtel Superstar II",
"     *.bin    Hemisphere Eclipse/Crescent",
"     *.stq    SkyTraq S1315F",
};
/* print help ----------------------------------------------------------------*/
static void printhelp(void)
{
    int i;
    for (i=0;i<sizeof(help)/sizeof(*help);i++) fprintf(stderr,"%s\n",help[i]);
    exit(0);
}
/* show message --------------------------------------------------------------*/
extern int showmsg(char *format, ...)
{
    va_list arg;
    va_start(arg,format); vfprintf(stderr,format,arg); va_end(arg);
    fprintf(stderr,"\r");
    return 0;
}
/* main ----------------------------------------------------------------------*/
static int convbin(int format, const rnxopt_t *rnxopt, char **file, char *dir)
{
    char ofile[1024],nfile[1024],gfile[1024],sfile[1024],work[1024],*p;
    
    if (file[1]) strcpy(ofile,file[1]);
    else {
        strcpy(ofile,file[0]);
        if ((p=strrchr(ofile,'.'))) strcpy(p,".obs"); else strcat(ofile,".obs");
    }
    if (file[2]) strcpy(nfile,file[2]);
    else {
        strcpy(nfile,file[0]);
        if ((p=strrchr(nfile,'.'))) strcpy(p,".nav"); else strcat(nfile,".nav");
    }
    if (file[3]) strcpy(gfile,file[3]);
    else {
        strcpy(gfile,file[0]);
        if ((p=strrchr(gfile,'.'))) strcpy(p,".gnav"); else strcat(gfile,".gnav");
    }
    if (file[4]) strcpy(sfile,file[4]);
    else {
        strcpy(sfile,file[0]);
        if ((p=strrchr(sfile,'.'))) strcpy(p,".sbs"); else strcat(sfile,".sbs");
    }
    if (dir&&*ofile) {
        if ((p=strrchr(ofile,PATHSEP))) strcpy(work,p+1); else strcpy(work,ofile);
        sprintf(ofile,"%s%c%s",dir,PATHSEP,work);
    }
    if (dir&&*nfile) {
        if ((p=strrchr(nfile,PATHSEP))) strcpy(work,p+1); else strcpy(work,nfile);
        sprintf(nfile,"%s%c%s",dir,PATHSEP,work);
    }
    if (dir&&*gfile) {
        if ((p=strrchr(gfile,PATHSEP))) strcpy(work,p+1); else strcpy(work,gfile);
        sprintf(gfile,"%s%c%s",dir,PATHSEP,work);
    }
    if (dir&&*sfile) {
        if ((p=strrchr(sfile,PATHSEP))) strcpy(work,p+1); else strcpy(work,sfile);
        sprintf(sfile,"%s%c%s",dir,PATHSEP,work);
    }
    fprintf(stderr,"input file: %s (%s)\n",file[0],rcvnamelist[format]);
    if (*ofile) fprintf(stderr,"rinex obs  : %s\n",ofile);
    if (*nfile) fprintf(stderr,"rinex nav  : %s\n",nfile);
    if (*gfile) fprintf(stderr,"rinex gnav : %s\n",gfile);
    if (*sfile) fprintf(stderr,"sbas msg   : %s\n",sfile);
    if (!convrnx(format,rnxopt,file[0],ofile,nfile,gfile,sfile)) {
        fprintf(stderr,"raw data conversion error\n");
        return -1;
    }
    fprintf(stderr,"\n");
    return 0;
}
/* main ----------------------------------------------------------------------*/
int main(int argc, char **argv)
{
    rnxopt_t rnxopt={{0}};
    double eps[]={1980,1,1,0,0,0},epe[]={2037,12,31,0,0,0};
    int i,format=-1,nf=2;
    char *rcv=NULL,*dir=NULL,*file[5]={0},*p;
    
    rnxopt.navsys =SYS_GPS|SYS_SBS;
    rnxopt.obstype=OBSTYPE_PR|OBSTYPE_CP;
    
    for (i=1;i<argc;i++) {
        if (!strcmp(argv[i],"-ts")&&i+2<argc) {
            sscanf(argv[++i],"%lf/%lf/%lf",eps,eps+1,eps+2);
            sscanf(argv[++i],"%lf:%lf:%lf",eps+3,eps+4,eps+5);
            rnxopt.ts=epoch2time(eps);
        }
        else if (!strcmp(argv[i],"-te")&&i+2<argc) {
            sscanf(argv[++i],"%lf/%lf/%lf",epe,epe+1,epe+2);
            sscanf(argv[++i],"%lf:%lf:%lf",epe+3,epe+4,epe+5);
            rnxopt.te=epoch2time(epe);
        }
        else if (!strcmp(argv[i],"-ti")&&i+1<argc) rnxopt.tint=atof(argv[++i]);
        else if (!strcmp(argv[i],"-f" )&&i+1<argc) nf=atoi(argv[++i]);
        else if (!strcmp(argv[i],"-od")) rnxopt.obstype|=OBSTYPE_DOP;
        else if (!strcmp(argv[i],"-os")) rnxopt.obstype|=OBSTYPE_SNR;
        else if (!strcmp(argv[i],"-r" )&&i+1<argc) rcv=argv[++i];
        else if (!strcmp(argv[i],"-d" )&&i+1<argc) dir=argv[++i];
        else if (!strcmp(argv[i],"-o" )&&i+1<argc) file[1]=argv[++i];
        else if (!strcmp(argv[i],"-n" )&&i+1<argc) file[2]=argv[++i];
        else if (!strcmp(argv[i],"-g" )&&i+1<argc) file[3]=argv[++i];
        else if (!strcmp(argv[i],"-s" )&&i+1<argc) file[4]=argv[++i];
        else if (!strncmp(argv[i],"-",1)) printhelp();
        else file[0]=argv[i];
    }
    if (!file[0]) {
        fprintf(stderr,"no input receiver log file\n");
        return -1;
    }
    if (rcv) {
        if      (!strcmp(rcv,"rtcm2")) format=STRFMT_RTCM2;
        else if (!strcmp(rcv,"rtcm3")) format=STRFMT_RTCM3;
        else if (!strcmp(rcv,"nov"  )) format=STRFMT_OEM4;
        else if (!strcmp(rcv,"oem3" )) format=STRFMT_OEM3;
        else if (!strcmp(rcv,"ubx"  )) format=STRFMT_UBX;
        else if (!strcmp(rcv,"ss2"  )) format=STRFMT_SS2;
        else if (!strcmp(rcv,"hemis")) format=STRFMT_CRES;
        else if (!strcmp(rcv,"stq"  )) format=STRFMT_STQ;
    }
    else if ((p=strrchr(file[0],'.'))) {
        if      (!strcmp(p,".rtcm2"))  format=STRFMT_RTCM2;
        else if (!strcmp(p,".rtcm3"))  format=STRFMT_RTCM3;
        else if (!strcmp(p,".gps"  ))  format=STRFMT_OEM4;
        else if (!strcmp(p,".ubx"  ))  format=STRFMT_UBX;
        else if (!strcmp(p,".log"  ))  format=STRFMT_SS2;
        else if (!strcmp(p,".bin"  ))  format=STRFMT_CRES;
        else if (!strcmp(p,".stq"  ))  format=STRFMT_STQ;
    }
    if (nf>=1) rnxopt.freqtype|=FREQTYPE_L1;
    if (nf>=2) rnxopt.freqtype|=FREQTYPE_L2;
    if (nf>=3) rnxopt.freqtype|=FREQTYPE_L5;
    if (nf>=4) rnxopt.freqtype|=FREQTYPE_L7;
    if (nf>=5) rnxopt.freqtype|=FREQTYPE_L6;
    
    if (format<0) {
        fprintf(stderr,"receiver type can not be recognized\n");
        return -1;
    }
    sprintf(rnxopt.prog,"%s %s",PRGNAME,VER_RTKLIB);
    sprintf(rnxopt.comment[0],"log: %-55.55s",file[0]);
    return convbin(format,&rnxopt,file,dir);
}
