/*------------------------------------------------------------------------------
* preceph.c : input precise ephemeris sp3 file
*
*          Copyright (C) 2007-2009 by T.TAKASU, All rights reserved.
*
* version : $Revision: 1.1 $ $Date: 2008/07/17 21:48:06 $
* history : 2009/01/18 1.0 new
*           2009/01/31 1.1 fix bug on numerical error to read sp3a ephemeris
*           2009/05/15 1.2 support glonass,galileo,qzs
*           2009/12/11 1.3 support wild-card expansion of file path
*-----------------------------------------------------------------------------*/
#include "rtklib.h"

static const char rcsid[]="$Id:$";

#define NMAX        10      /* order of polynomial interpolation */
#define MAXDTE      900.0   /* max time difference to ephemeris time (s) */

/* satellite code to satellite system ----------------------------------------*/
static int code2sys(char code)
{
    if (code=='G'||code==' ') return SYS_GPS;
    if (code=='R') return SYS_GLO;
    if (code=='E') return SYS_GAL; /* extension, no sp3c */
    if (code=='Q') return SYS_QZS; /* extension, no sp3c */
    return SYS_NONE;
}
/* read sp3 header -----------------------------------------------------------*/
static int readsp3h(FILE *fp, gtime_t *time, char *type, int *sats,
                    double *bfact, char *tsys)
{
    int i,j,k=0,ns=0,sys,prn,sat;
    char buff[1024];
    
    trace(3,"readsp3h:\n");
    
    for (i=0;i<22;i++) {
        if (!fgets(buff,sizeof(buff),fp)) break;
        
        if (i==0) {
            *type=buff[2];
            if (str2time(buff,3,28,time)) return 0;
        }
        else if (2<=i&&i<=6) {
            if (i==2) {
                ns=(int)str2num(buff,4,2);
            }
            for (j=0;j<17&&k<ns;j++) {
                sys=code2sys(buff[9+3*j]);
                prn=(int)str2num(buff,10+3*j,2);
                
                if (!(sat=satno(sys,prn))) {
                    trace(2,"readsp3h: satellite number error: sys=%d prn=%d\n",sys,prn);
                }
                sats[k++]=sat;
            }
        }
        else if (i==12) {
            strncpy(tsys,buff+9,3); tsys[3]='\0';
        }
        else if (i==14) {
            bfact[0]=str2num(buff,3,10);
            bfact[1]=str2num(buff,14,12);
        }
    }
    return ns;
}
/* read sp3 body -------------------------------------------------------------*/
static void readsp3b(FILE *fp, char type, int *sats, int ns, double *bfact,
                     char *tsys, int index, peph_t **peph, int *n, int *nmax)
{
    double std,ep[6],val,fact[]={1E3,1E3,1E3,1E-6};
    int i,j,sat,sys,prn;
    char buff[1024];
    peph_t peph0={{0}},*p;
    
    trace(3,"readsp3b: index=%d\n",index);
    
    while (fgets(buff,sizeof(buff),fp)!=NULL) {
        if (!strncmp(buff,"EOF",3)) break;
        
        if (sscanf(buff+2,"%lf %lf %lf %lf %lf %lf",ep,ep+1,ep+2,ep+3,ep+4,ep+5)<6) {
            continue;
        }
        if (*nmax==0||*n>=*nmax) {
            *nmax+=1024;
            if (!(*peph=(peph_t *)realloc(*peph,sizeof(peph_t)*(*nmax)))) {
                trace(1,"readsp3b: malloc error: nmax=%d\n",*nmax);
                *n=*nmax=0;
                return;
            }
        }
        p=*peph+(*n);
        *p=peph0;
        p->time=epoch2time(ep);
        if (!strcmp(tsys,"UTC")) p->time=utc2gpst(p->time); /* utc->gpst */
        p->index=index;
        for (i=0;i<ns*(type=='P'?1:2);i++) {
            if (!fgets(buff,sizeof(buff),fp)) break;
            if (buff[0]!='P') continue;
            sys=code2sys(buff[1]);
            prn=(int)str2num(buff,2,2);
            if (!(sat=satno(sys,prn))) {
                trace(2,"readsp3b: satellite number error: sys=%d prn=%d\n",sys,prn);
                continue;
            }
            for (j=0;j<4;j++) {
                if ((val=str2num(buff,j*14+4,14))==0.0) continue;
                if (fabs(val-999999.999999)<1E-6) continue;
                p->pos[sat-1][j]=val*fact[j];
                if (j<3) {
                    std=str2num(buff,j*3+61,2);
                    p->std[sat-1][j]=bfact[0]<=0.0?0.0:pow(bfact[0],std)*1E-3;
                }
                else {
                    std=str2num(buff,j*3+61,3);
                    p->std[sat-1][j]=bfact[1]<=0.0?0.0:pow(bfact[1],std)*1E-12;
                }
            }
        }
        (*n)++;
    }
}
/* compare precise ephemerides -----------------------------------------------*/
static int cmppeph(const void *p1, const void *p2)
{
    peph_t *q1=(peph_t *)p1,*q2=(peph_t *)p2;
    double tt=timediff(q1->time,q2->time);
	return tt<-1E-12?-1:(tt>1E-12?1:q1->index<q2->index);
}
/* sort and combine precise ephemris -----------------------------------------*/
static int cmbpeph(peph_t *peph, int n)
{
    double tt;
    int i,j,k,sat;
    
    trace(3,"cmbeph: n=%d\n",n);
    
    qsort(peph,n,sizeof(peph_t),cmppeph);
    
    for (i=0,j=1;j<n;j++) {
        tt=timediff(peph[i].time,peph[j].time);
        
        if (fabs(tt)<=1E-12) { /* same epoch */
            
            /* copy later field to previous (precedence of later) */
            for (sat=1;sat<=MAXSAT;sat++) {
                if (norm(peph[j].pos[sat-1],4)<=0.0) continue;
                for (k=0;k<4;k++) peph[i].pos[sat-1][k]=peph[j].pos[sat-1][k];
                for (k=0;k<4;k++) peph[i].std[sat-1][k]=peph[j].std[sat-1][k];
            }
        }
        else if (++i<j) {
           peph[i]=peph[j];
        }
    }
    return i+1;
}
/* read sp3 files --------------------------------------------------------------
* read sp3 precise ephmeris files
* args   : char   *file       I   sp3 precise ephemeris files
*                                 (wind-card * is expanded)
*          peph_t **peph      O   precise ephemeris
* return : number of precise epheremeris (0:error)
*-----------------------------------------------------------------------------*/
extern int readsp3(const char *file, peph_t **peph)
{
    FILE *fp;
    gtime_t time;
    double bfact[2];
    char type=' ',tsys[4]="",*efiles[MAXEXFILE]={0};
    int i,n=0,nmax=0,nfile,ns,sats[MAXSAT]={0};
    
    trace(3,"readsp3: file=%s\n",file);
    
    *peph=NULL;
    
    for (i=0;i<MAXEXFILE;i++) {
        if (!(efiles[i]=(char *)malloc(1024))) {
            for (i--;i>=0;i--) free(efiles[i]);
            return 0;
        }
    }
    /* expand wild card in file path */
    nfile=expath(file,efiles,MAXEXFILE);
    
    for (i=0;i<nfile;i++) {
        if (!(fp=fopen(efiles[i],"r"))) {
            trace(1,"readsp3: file open error: %s\n",efiles[i]);
            continue;
        }
        /* read sp3 file */
        if ((ns=readsp3h(fp,&time,&type,sats,bfact,tsys))>0) {
            readsp3b(fp,type,sats,ns,bfact,tsys,i,peph,&n,&nmax);
        }
        fclose(fp);
    }
    for (i=0;i<MAXEXFILE;i++) free(efiles[i]);
    
    /* sort and combine precise ephemerides */
    if (n>0) n=cmbpeph(*peph,n);
    
    return n;
}
/* read satellite antenna phase center position --------------------------------
* read satellite antenna phase center position
* args   : char   *file       I   antenna pcv file (antex)
*          gtime_t time       I   time
*          double sap[MAXSAT][3] O satellite antenna phase center pos (m)
*                                 sap[sat-1][] : satellite sat phase center (m)
*                                 {x,y,z} in satellite fixed coordinates
* return : status (1:ok,0:file open error)
*-----------------------------------------------------------------------------*/
extern int readsap(const char *file, gtime_t time, double sap[][3])
{
    FILE *fp;
    const double posp[2][3]={{0.279,0.0,1.023},{0.0,0.0,0.0}};
    const double epswp[]={2006,11,5,0,0,0}; /* igs antenna model swap */
    gtime_t ts={0},te={0},t0={0},tswp;
    double pos[3];
    char buff[1024];
    int i,j,prn=-1,sys=0,sat,block=0;
    
    trace(3,"readsap: file=%s\n",file);
    
    if (!(fp=fopen(file,"r"))) {
        trace(1,"readsap: file open error %s\n",file);
        return 0;
    }
    tswp=epoch2time(epswp);
    for (i=0;i<MAXSAT;i++) for (j=0;j<3;j++) sap[i][j]=0.0;
    
    while (fgets(buff,sizeof(buff),fp)) {
        if (strlen(buff)<60) continue;
        if (strstr(buff+60,"END OF ANTENNA")) {
            prn=-1; ts=t0; te=t0;
        }
        if (strstr(buff+60,"START OF ANTENNA")) prn=0;
        if (prn<0) continue;
        if (strstr(buff+60,"TYPE / SERIAL NO")) {
            sys=code2sys(buff[20]);
            prn=(int)str2num(buff,21,2);
            if (strstr(buff,"BLOCK IIA")==buff) block=0; else block=1;
            if (strncmp(buff+23,"    ",4)) sys=SYS_NONE; /* receiver ant */
        }
        else if (strstr(buff+60,"VALID FROM")) {
            if (!str2time(buff,0,43,&ts)) continue;
        }
        else if (strstr(buff+60,"VALID UNTIL")) {
            if (!str2time(buff,0,43,&te)) continue;
        }
        else if (strstr(buff+60,"NORTH / EAST / UP")) {
            if (prn==0) continue; /* receiver */
            if (sscanf(buff,"%lf %lf %lf",pos,pos+1,pos+2)<3) continue;
            if (ts.time!=0&&timediff(ts,time)>0.0) continue;
            if (te.time!=0&&timediff(te,time)<0.0) continue;
            if (!(sat=satno(sys,prn))) {
                trace(2,"readsap: satellite number error: sys=%d prn=%d\n",sys,prn);
                continue;
            }
            if (timediff(time,tswp)>=0.0) {
                for (i=0;i<3;i++) sap[sat-1][i]=pos[i]*1E-3;
            }
            else if (sys==SYS_GPS) {
                for (i=0;i<3;i++) sap[sat-1][i]=posp[block][i];
            }
        }
    }
    fclose(fp);
    return 1;
}
/* polynomial interpolation by neville's algorithm ---------------------------*/
static double interppol(const double *x, double *y, int n)
{
    int i,j;
    
    for (j=1;j<n;j++) {
        for (i=0;i<n-j;i++) {
            y[i]=(x[i+j]*y[i]-x[i]*y[i+1])/(x[i+j]-x[i]);
        }
    }
    return y[0];
}
/* satellite precise ephemeris to satellite position/clock-bias ----------------
* compute satellite position with precise ephemeris
* args   : gtime_t time       I   time
*          int     sat        I   satellite number
*          peph_t *eph        I   precise ephemeris records (sorted by time)
*          int    neph        I   number of precise ephemeris
*          double *rs         O   satellite position (ecef) (m)
*          double *dts        O   satellite clock-bias (s)
* return : status (1:ok,0:error or data outage)
*-----------------------------------------------------------------------------*/
extern int eph2posp(gtime_t time, int sat, peph_t *eph, int neph, double *rs,
                    double *dts)
{
    double x[NMAX+1],y[NMAX+1];
    int i,j,k,n,m;
    
    trace(4,"eph2posp: sat=%d neph=%d\n",sat,neph);
    
    rs[0]=rs[1]=rs[2]=*dts=0.0;
    
    if (neph<NMAX+1||sat<1||MAXSAT<sat) return 0;
    
    if (timediff(time,eph[0].time)<-MAXDTE||
        timediff(time,eph[neph-1].time)>MAXDTE) return 0;
    
    /* binary search */
    for (i=0,j=neph-1;i<j;) {
        k=(i+j)/2;
        if (timediff(eph[k].time,time)<0.0) i=k+1; else j=k;
    }
    n=i; m=n-(NMAX+1)/2;
    if (m<0) m=0; else if (m+NMAX>=neph) m=neph-NMAX-1;
    
    for (i=0;i<=NMAX;i++) {
        x[i]=timediff(eph[m+i].time,time);
        
        /* ephemeris outage */
        if (norm(eph[m+i].pos[sat-1],3)<=0.0) return 0;
    }
    /* polynomial interpolation for orbit */
    for (i=0;i<3;i++) {
        for (j=0;j<=NMAX;j++) y[j]=eph[m+j].pos[sat-1][i];
        rs[i]=interppol(x,y,NMAX+1);
    }
    /* linear interpolation for clock */
    if (n<=0) {
        *dts=eph[0].pos[sat-1][3];
        return *dts!=0.0;
    }
    for (j=0;j<2;j++) {
        x[j]=timediff(eph[n+j-1].time,time);
        if ((y[j]=eph[n+j-1].pos[sat-1][3])==0.0) return 0;
    }
    *dts=interppol(x,y,2);
    return 1;
}
/* satellite positions/clock-biases with precise ephemeris ---------------------
* compute satellite positions and clock-biases with precise ephemeris
* args   : gtime_t time     I   time to search ephemeris
*          obsd_t *obs      I   observation data records
*          int     n        I   number of observation data records
*          nav_t  *nav      I   navigation messages
*          double *rs       O   satellilte positions (ecef) (m)
*          double *dts      O   satellilte clock-bias (sec)
* return : none
* notes  : rs[(0:2)+i*3],dts[i] = obs[i] satellite position/clock-bias
*          if no navigation messages, set 0.0 to rs[], dts[]
*-----------------------------------------------------------------------------*/
extern void satposp(gtime_t time, const obsd_t *obs, int n, const nav_t *nav,
                    double *rs, double *dts)
{
    double rsp[3],rss[3],vs[3],tt,r;
    int i,j,sat;
    
    trace(4,"satposp: n=%d\n",n);
    
    for (i=0;i<n&&i<MAXOBS;i++) {
        rs[i*3]=rs[1+i*3]=rs[2+i*3]=dts[i]=0.0;
        if (obs[i].P[0]==0.0) continue;
        sat=obs[i].sat;
        time=timeadd(obs[i].time,-obs[i].P[0]/CLIGHT);
        if (!eph2posp(time,sat,nav->peph,nav->ne,rsp,&tt)) continue;
        time=timeadd(time,-tt);
        if (!eph2posp(time,sat,nav->peph,nav->ne,rss,dts+i)) continue;
        r=norm(rss,3);
        for (j=0;j<3;j++) {
            rs[j+i*3]=rss[j]-nav->sap[sat-1][2]*rss[j]/r; /* ant z-pos */
            vs[j]=-(rss[j]-rsp[j])/tt;
        }
        dts[i]-=2.0*dot(rs+i*3,vs,3)/CLIGHT/CLIGHT; /* relativ */
    }
}
