/*------------------------------------------------------------------------------
* rtkpos.c : realtime kinematic positioning
*
*          Copyright (C) 2007-2009 by T.TAKASU, All rights reserved.
*
* version : $Revision: 1.1 $ $Date: 2008/07/17 21:48:06 $
* history : 2007/01/12 1.0 new
*           2007/03/13 1.1 add slip detection by LLI flag
*           2007/04/18 1.2 add antenna pcv correction
*                          change rtkpos argin
*           2008/07/18 1.3 refactored
*           2009/01/02 1.4 modify rtk positioning api
*           2009/03/09 1.5 support glonass, gallileo and qzs
*           2009/08/27 1.6 fix bug on numerical exception
*           2009/09/03 1.7 add check of valid satellite number
*                          add check time sync for moving-base
*           2009/11/23 1.8 add api rtkopenstat(),rtkclosestat()
*                          add receiver h/w bias estimation
*                          add solution status output
*-----------------------------------------------------------------------------*/
#include <stdarg.h>
#include "rtklib.h"

static const char rcsid[]="$Id: rtkpos.c,v 1.1 2008/07/17 21:48:06 ttaka Exp $";

/* constants/macros ----------------------------------------------------------*/

#define VAR_POS     900.0    /* initial variance of receiver pos (m^2) */
#define VAR_VEL     100.0    /* initial variance of receiver vel ((m/s)^2) */
#define VAR_ACC     100.0    /* initial variance of receiver acc ((m/ss)^2) */
#define VAR_HWBIAS  1.0      /* initial variance of h/w bias ((m/MHz)^2) */
#define VAR_FIXAMB  1E-6     /* constraint to fixed ambiguity (cycle^2) */
#define PRN_HWBIAS  1E-6     /* process noise of h/w bias (m/MHz/sqrt(s)) */
#define GAP_RESION  60       /* gap to reset ionosphere parameters (epochs) */
#define NFIXTOHOLD  10       /* number of continuous fixes to hold ambiguity */
#define MAXACC      30.0     /* max accel for doppler slip detection (m/s^2) */

#define TTOL_MOVEB  0.2      /* time sync tolerance for moving-baseline (s) */

#define SQR(x)      ((x)*(x))
#define SQRT(x)     ((x)<=0.0?0.0:sqrt(x))
#define MIN(x,y)    ((x)<=(y)?(x):(y))
#define NP(opt)     (!(opt)->dynamics?3:9) /* number of solution */
#define II(s,opt)   (NP(opt)+(s)-1)
                             /* state index of ionos (s:satellite no) */
#define IT(r,opt)   (II(1,opt)+((opt)->estiono?MAXSAT:0)+(r))
                             /* state index of tropos (r:0=rov,1:ref) */
#define IL(f,opt)   (IT(0,opt)+((opt)->esttrop?2:0)+(f))
                             /* state index of receiver h/w bias */
#define NR(opt)     (IL(0,opt)+((opt)->glomodear==2?NFREQGLO:0))
                             /* number of real prameters */
#define IB(s,f,opt) (NR(opt)+MAXSAT*(f)+(s)-1)
                             /* state index of phase bias */

/* global variables ----------------------------------------------------------*/
static int statlevel=0;          /* rtk status output level (0:off) */
static FILE *fp_stat=NULL;       /* rtk status file pointer */

/* open solution status file ---------------------------------------------------
* open solution status file and set output level
* args   : char     *file   I   rtk status file
*          int      level   I   rtk status level (0: off)
* return : status (1:ok,0:error)
* output : solution status file record format
*
*   $POS,week,tow,stat,posx,posy,posz,posxf,posyf,poszf
*          week/tow : gps week no/time of week (s)
*          stat     : solution status
*          posx/posy/posz    : position x/y/z ecef (m) float
*          posxf/posyf/poszf : position x/y/z ecef (m) fixed
*
*   $VELACC,week,tow,stat,vele,veln,velu,acce,accn,accu,velef,velnf,veluf,accef,accnf,accuf
*          week/tow : gps week no/time of week (s)
*          stat     : solution status
*          vele/veln/velu    : velocity e/n/u (m/s) float
*          acce/accn/accu    : acceleration e/n/u (m/s^2) float
*          velef/velnf/veluf : velocity e/n/u (m/s) fixed
*          accef/accnf/accuf : acceleration e/n/u (m/s^2) fixed
*
*   $ION,week,tow,stat,sat,az,el,ion,ion-fixed
*          week/tow : gps week no/time of week (s)
*          stat     : solution status
*          sat      : satellite id
*          az/el    : azimuth/elevation angle(deg)
*          ion      : vertical ionospheric delay L1 (m) float
*          ion-fixed: vertical ionospheric delay L1 (m) fixed
*
*   $TROP,week,tow,stat,rcv,ztd,ztdf
*          week/tow : gps week no/time of week (s)
*          stat     : solution status
*          rcv      : receiver (1:rover,2:base station)
*          ztd      : zenith total delay (m) float
*          ztdf     : zenith total delay (m) fixed
*
*   $HWBIAS,week,tow,stat,frq,bias,biasf
*          week/tow : gps week no/time of week (s)
*          stat     : solution status
*          frq      : frequency (1:L1,2:L2,...)
*          bias     : h/w bias coefficient (m/MHz) float
*          biasf    : h/w bias coefficient (m/MHz) fixed
*
*   $SAT,week,tow,sat,frq,az,el,resp,resc,vsat,snr,fix,slip,lock,outc,slipc,rejc
*          week/tow : gps week no/time of week (s)
*          sat/frq  : satellite id/frequency (1:L1,2:L2,...)
*          az/el    : azimuth/elevation angle (deg)
*          resp     : pseudorange residual (m)
*          resc     : carrier-phase residual (m)
*          vsat     : valid data flag (0:invalid,1:valid)
*          snr      : signal strength (dbHz)
*          fix      : ambiguity flag  (0:no data,1:float,2:fixed,3:hold)
*          slip     : cycle-slip flag (bit1:slip,bit2:parity unknown)
*          lock     : carrier-lock count
*          outc     : data outage count
*          slipc    : cycle-slip count
*          rejc     : data reject (outlier) count
*
*-----------------------------------------------------------------------------*/
extern int rtkopenstat(const char *file, int level)
{
    trace(3,"rtkopenstat : file=%s level=%d\n",file,level);
    
    if (level<=0) return 0;
    if (!(fp_stat=fopen(file,"w"))) {
        trace(1,"rtkopenstat: file open error file=%s\n",file);
        return 0;
    }
    statlevel=level;
    return 1;
}
/* close solution status file --------------------------------------------------
* close solution status file
* args   : none
* return : none
*-----------------------------------------------------------------------------*/
extern void rtkclosestat(void)
{
    trace(3,"rtkclosestat :\n");
    
    if (fp_stat) fclose(fp_stat);
    fp_stat=NULL;
    statlevel=0;
}
/* output solution status ----------------------------------------------------*/
static void outsolstat(rtk_t *rtk, int stat)
{
    double tow,pos[3],vel[3],acc[3],vela[3],acca[3];
    int i,j,week,est,nfreq;
    char id[32];
    
    if (statlevel<=0||!fp_stat) return;
    
    est=rtk->opt.mode>=PMODE_DGPS;
    nfreq=est?rtk->opt.nf:1;
    tow=time2gpst(rtk->time[0],&week);
    
    /* receiver position */
    if (est) {
        fprintf(fp_stat,"$POS,%d,%.3f,%d,%.4f,%.4f,%.4f,%.4f,%.4f,%.4f\n",week,tow,
                stat,rtk->x[0],rtk->x[1],rtk->x[2],rtk->xa[0],rtk->xa[1],rtk->xa[2]);
    }
    /* receiver velocy and acceleration */
    if (est&&rtk->opt.dynamics) {
        ecef2pos(rtk->rr[0],pos);
        ecef2enu(pos,rtk->x+3,vel);
        ecef2enu(pos,rtk->x+6,acc);
        ecef2enu(pos,rtk->xa+3,vela);
        ecef2enu(pos,rtk->xa+6,acca);
        fprintf(fp_stat,"$VELACC,%d,%.3f,%d,%.4f,%.4f,%.4f,%.5f,%.5f,%.5f,%.4f,%.4f,%.4f,%.5f,%.5f,%.5f\n",
                week,tow,stat,vel[0],vel[1],vel[2],acc[0],acc[1],acc[2],
                vela[0],vela[1],vela[2],acca[0],acca[1],acca[2]);
    }
    /* ionospheric parameters */
    if (est&&rtk->opt.estiono) {
        for (i=0;i<MAXSAT;i++) {
            if (!rtk->vs[0][i]) continue;
            satno2id(i+1,id);
            j=II(i+1,&rtk->opt);
            fprintf(fp_stat,"$ION,%d,%.3f,%d,%s,%.1f,%.1f,%.4f,%.4f\n",week,tow,stat,
                    id,rtk->az[0][i]*R2D,rtk->el[0][i]*R2D,rtk->x[j],rtk->xa[j]);
        }
    }
    /* tropospheric parameters */
    if (est&&rtk->opt.esttrop) {
        for (i=0;i<2;i++) {
            j=IT(i,&rtk->opt);
            fprintf(fp_stat,"$TROP,%d,%.3f,%d,%d,%.4f,%.4f\n",week,tow,stat,i+1,
                    rtk->x[j],rtk->xa[j]);
        }
    }
    /* receiver h/w bias */
    if (est&&rtk->opt.glomodear==2) {
        for (i=0;i<nfreq;i++) {
            j=IL(i,&rtk->opt);
            fprintf(fp_stat,"$HWBIAS,%d,%.3f,%d,%d,%.4f,%.4f\n",week,tow,stat,i+1,
                    rtk->x[j],rtk->xa[j]);
        }
    }
    if (stat==SOLQ_NONE||statlevel<=1) return;
    
    /* residuals and status */
    for (i=0;i<MAXSAT;i++) {
        if (!rtk->vs[0][i]) continue;
        satno2id(i+1,id);
        for (j=0;j<nfreq;j++) {
            fprintf(fp_stat,"$SAT,%d,%.3f,%s,%d,%.1f,%.1f,%.4f,%.4f,%d,%d,%d,%d,%d,%d,%d,%d\n",
                    week,tow,id,j+1,rtk->az[0][i]*R2D,rtk->el[0][i]*R2D,
                    rtk->resp[i][j],rtk->resc[i][j],rtk->vsat[i][j],
                    rtk->snr [i][j],rtk->fix [i][j],rtk->slip[i][j]&3,
                    rtk->lock[i][j],rtk->outc[i][j],rtk->slipc[i][j],
                    rtk->rejc[i][j]);
        }
    }
}
/* save error message --------------------------------------------------------*/
static void errmsg(rtk_t *rtk, const char *format, ...)
{
    char buff[256];
    int n;
    va_list ap;
    n=sprintf(buff,"%s: ",rtk->tstr);
    va_start(ap,format);
    n+=vsprintf(buff+n,format,ap);
    va_end(ap);
    n=n<MAXERRMSG-rtk->neb?n:MAXERRMSG-rtk->neb;
    memcpy(rtk->errbuf+rtk->neb,buff,n);
    rtk->neb+=n;
}
/* single-differenced observable ---------------------------------------------*/
static double sdobs(const obsd_t *obs, int i, int j, int f)
{
    double pi=f<NFREQ?obs[i].L[f]:obs[i].P[f-NFREQ];
    double pj=f<NFREQ?obs[j].L[f]:obs[j].P[f-NFREQ];
    return pi==0.0||pj==0.0?0.0:pi-pj;
}
/* single-differenced geometry-free linear combination of phase --------------*/
static double gfobs(const obsd_t *obs, int i, int j, double *lams)
{
    double pi=sdobs(obs,i,j,0)*lams[0],pj=sdobs(obs,i,j,1)*lams[1];
    return pi==0.0||pj==0.0?0.0:pi-pj;
}
/* single-differenced measurement error variance -----------------------------*/
static double varerr(int sys, double el, double bl, double dt, int f,
                     const prcopt_t *opt)
{
    double a=opt->err[1],b=opt->err[2]/sin(el),c=opt->err[3]*bl/1E4;
    double d=CLIGHT*opt->sclkstab*dt,fact=f<opt->nf?1.0:opt->err[0];
    
    /* system factor affects to both of code and phase error (v.2.3.0) */
#if 1
    fact*=sys==SYS_GLO?EFACT_GLO:(sys==SYS_SBS?EFACT_SBS:EFACT_GPS);
#else
    if (f>=opt->nf) {
        fact*=sys==SYS_GLO?EFACT_GLO:(sys==SYS_SBS?EFACT_SBS:EFACT_GPS);
    }
#endif
    return 2.0*(a*a*fact*fact+b*b*fact*fact+c*c)+d*d;
}
/* baseline length -----------------------------------------------------------*/
static double baseline(const double *ru, const double *rb, double *dr)
{
    dr[0]=ru[0]-rb[0]; dr[1]=ru[1]-rb[1]; dr[2]=ru[2]-rb[2];
    return norm(dr,3);
}
/* initialize state and covariance -------------------------------------------*/
static void initx(rtk_t *rtk, double xi, double var, int i)
{
    int j;
    rtk->x[i]=xi;
    for (j=0;j<rtk->nx;j++) {
        rtk->P[i+j*rtk->nx]=rtk->P[j+i*rtk->nx]=i==j?var:0.0;
    }
}
/* select common satellites between rover and reference station --------------*/
static int selsat(const obsd_t *obs, double *azel, int nu, int nr,
                  const prcopt_t *opt, int *sat, int *iu, int *ir)
{
    int i,j,k=0;
    
    trace(3,"selsat : nu=%d nr=%d\n",nu,nr);
    
    for (i=0,j=nu;i<nu&&j<nu+nr;i++,j++) {
        if      (obs[i].sat<obs[j].sat) j--;
        else if (obs[i].sat>obs[j].sat) i--;
        else if (azel[1+j*2]>=opt->elmin) { /* elevation at base station */
            sat[k]=obs[i].sat; iu[k]=i; ir[k++]=j;
            trace(4,"(%2d) sat=%3d iu=%2d ir=%2d\n",k-1,obs[i].sat,i,j);
        }
    }
    return k;
}
/* temporal update of position/velocity/acceleration -------------------------*/
static void udpos(rtk_t *rtk, double tt)
{
    double *F,*FP,*xp,pos[3],Q[9]={0},Qv[9],var=0.0;
    int i,j;
    
    trace(3,"udpos : tt=%.3f\n",tt);
    
    /* fixed mode */
    if (rtk->opt.mode==PMODE_FIXED) {
        for (i=0;i<3;i++) initx(rtk,rtk->opt.ru[i],1E-8,i);
        return;
    }
    /* initialize position for first epoch */
    if (norm(rtk->x,3)<=0.0) {
        for (i=0;i<3;i++) initx(rtk,rtk->rr[0][i],VAR_POS,i);
        if (rtk->opt.dynamics) {
            for (i=3;i<6;i++) initx(rtk,rtk->rr[0][i],VAR_VEL,i);
            for (i=6;i<9;i++) initx(rtk,1E-6,VAR_ACC,i);
        }
    }
    /* static mode */
    if (rtk->opt.mode==PMODE_STATIC) return;
    
    /* kinmatic mode without dynamics */
    if (!rtk->opt.dynamics) {
        for (i=0;i<3;i++) initx(rtk,rtk->rr[0][i],VAR_POS,i);
        return;
    }
    /* check variance of estimated postion */
    for (i=0;i<3;i++) var+=rtk->P[i+i*rtk->nx]; var/=3.0;
    
    if (var>VAR_POS) {
        /* reset position with large variance */
        for (i=0;i<3;i++) initx(rtk,rtk->rr[0][i],VAR_POS,i);
        for (i=3;i<6;i++) initx(rtk,rtk->rr[0][i],VAR_VEL,i);
        for (i=6;i<9;i++) initx(rtk,1E-6,VAR_ACC,i);
        trace(2,"uppos  : reset position due to large variance (%s var=%.3f)\n",
             rtk->tstr,var);
        return;
    }
    /* state transition of position/velocity/acceleration */
    F=eye(rtk->nx); FP=mat(rtk->nx,rtk->nx); xp=mat(rtk->nx,1);
    
    for (i=0;i<6;i++) {
        F[i+(i+3)*rtk->nx]=tt;
    }
    /* x=F*x, P=F*P*F+Q */
    matmul("NN",rtk->nx,1,rtk->nx,1.0,F,rtk->x,0.0,xp);
    matcpy(rtk->x,xp,rtk->nx,1);
    matmul("NN",rtk->nx,rtk->nx,rtk->nx,1.0,F,rtk->P,0.0,FP);
    matmul("NT",rtk->nx,rtk->nx,rtk->nx,1.0,FP,F,0.0,rtk->P);
    
    /* process noise added to only acceleration */
    Q[0]=Q[4]=SQR(rtk->opt.prn[3]); Q[8]=SQR(rtk->opt.prn[4]);
    ecef2pos(rtk->x,pos);
    covecef(pos,Q,Qv);
    for (i=0;i<3;i++) for (j=0;j<3;j++) {
        rtk->P[i+6+(j+6)*rtk->nx]+=Qv[i+j*3];
    }
    free(F); free(FP); free(xp);
}
/* temporal update of ionospheric parameters ---------------------------------*/
static void udion(rtk_t *rtk, double tt, double bl, const int *sat, int ns)
{
    double el,fact;
    int i,j;
    
    trace(3,"udion: tt=%.1f bl=%.0f ns=%d\n",tt,bl,ns);
    
    for (i=1;i<=MAXSAT;i++) {
        j=II(i,&rtk->opt);
        if (rtk->x[j]!=0.0&&
            rtk->outc[i-1][0]>GAP_RESION&&rtk->outc[i-1][1]>GAP_RESION)
            rtk->x[j]=0.0;
    }
    for (i=0;i<ns;i++) {
        j=II(sat[i],&rtk->opt);
        
        if (rtk->x[j]==0.0) {
            initx(rtk,1E-6,SQR(rtk->opt.std[1]*bl/1E4),j);
        }
        else {
            /* elevation dependent factor of process noise */
            el=rtk->el[0][sat[i]-1];
            if (el<5.0*D2R) el=5.0*D2R;
            fact=1.0/sin(el);
            
            rtk->P[j+j*rtk->nx]+=SQR(rtk->opt.prn[1]*bl/1E4*fact)*tt;
        }
    }
}
/* temporal update of tropospheric parameters --------------------------------*/
static void udtrop(rtk_t *rtk, double tt)
{
    int i,j;
    
    trace(3,"udtrop: tt=%.1f\n",tt);
    
    for (i=0;i<2;i++) {
        j=IT(i,&rtk->opt);
        
        if (rtk->x[j]==0.0) {
            initx(rtk,1E-6,SQR(rtk->opt.std[2]),j);
        }
        else {
            rtk->P[j+j*rtk->nx]+=SQR(rtk->opt.prn[2])*tt;
        }
    }
}
/* temporal update of receiver h/w biases ------------------------------------*/
static void udrcvbias(rtk_t *rtk, double tt)
{
    int i,j;
    
    trace(3,"udrcvbias: tt=%.1f\n",tt);
    
    for (i=0;i<NFREQGLO;i++) {
        j=IL(i,&rtk->opt);
        
        if (rtk->x[j]==0.0) {
            initx(rtk,1E-6,VAR_HWBIAS,j);
        }
        /* hold to fixed solution */
        else if (rtk->nfix>=NFIXTOHOLD&&rtk->ambratio>rtk->opt.thresar) {
            initx(rtk,rtk->xa[j],rtk->Pa[j+j*rtk->na],j);
        }
        else {
            rtk->P[j+j*rtk->nx]+=SQR(PRN_HWBIAS)*tt;
        }
    }
}
/* detect cycle slip by LLI --------------------------------------------------*/
static void detslp_lli(rtk_t *rtk, const obsd_t *obs, int i, int rcv)
{
    unsigned char slip,LLI1,LLI2,LLI;
    int f,sat=obs[i].sat;
    char tstr[32];
    
    time2str(obs[i].time,tstr,2);
    
    for (f=0;f<rtk->opt.nf;f++) {
        
        if (obs[i].L[f]==0.0) continue;
        
        /* restore previous LLI */
        LLI1=(rtk->slip[sat-1][f]>>6)&3;
        LLI2=(rtk->slip[sat-1][f]>>4)&3;
        LLI=rcv==1?LLI1:LLI2;
        
        /* detect slip by cycle slip flag */
        slip=(rtk->slip[sat-1][f]|obs[i].LLI[f])&3;
        
        if (obs[i].LLI[f]&1) {
            errmsg(rtk,"slip detected (sat=%2d rcv=%d LLI%d=%x)\n",
                   sat,rcv,f+1,obs[i].LLI[f]);
            trace(2,"detslp : slip detected (%s sat=%2d rcv=%d LLI%d=%x)\n",
                  tstr+11,sat,rcv,f+1,obs[i].LLI[f]);
        }
        /* detect slip by parity unknown flag transition */
        if (((LLI&2)&&!(obs[i].LLI[f]&2))||(!(LLI&2)&&(obs[i].LLI[f]&2))) {
            errmsg(rtk,"slip detected (sat=%2d rcv=%d LLI%d=%x->%x)\n",
                   sat,rcv,f+1,LLI,obs[i].LLI[f]);
            trace(2,"detslp : slip detected (%s sat=%2d rcv=%d LLI%d=%x->%x)\n",
                  tstr+11,sat,rcv,f+1,LLI,obs[i].LLI[f]);
            slip|=1;
        }
        /* save current LLI and slip flag */
        if (rcv==1) rtk->slip[sat-1][f]=(obs[i].LLI[f]<<6)|(LLI2<<4)|slip;
        else        rtk->slip[sat-1][f]=(obs[i].LLI[f]<<4)|(LLI1<<6)|slip;
    }
}
/* detect cycle slip by geometry free phase jump -----------------------------*/
static void detslp_gf(rtk_t *rtk, const obsd_t *obs, int i, int j)
{
    int f,sat=obs[i].sat;
    double g0,g1;
    
    if (rtk->opt.nf<=1||(g1=gfobs(obs,i,j,rtk->lams[sat-1]))==0.0) return;
    
    g0=rtk->gf[sat-1]; rtk->gf[sat-1]=g1;
        
    if (g0!=0.0&&fabs(g1-g0)>rtk->opt.thresslip) {
        
        for (f=0;f<rtk->opt.nf;f++) rtk->slip[sat-1][f]|=1;
        
        errmsg(rtk,"slip detected (sat=%3d GF=%.3f %.3f)\n",sat,g0,g1);
        trace(2,"detslp : slip detected (%s sat=%3d GF=%.3f %.3f)\n",
              rtk->tstr,sat,g0,g1);
    }
}
/* detect cycle slip by doppler and phase difference -------------------------*/
static void detslp_dop(rtk_t *rtk, const obsd_t *obs, int i, int rcv)
{
    /* detection with doppler disabled because of clock-jump issue (v.2.3.0) */
#if 0
    int f,sat=obs[i].sat;
    double tt,dph,dpt,lam,thres;
    
    for (f=0;f<rtk->opt.nf;f++) {
        if (obs[i].L[f]==0.0||obs[i].D[f]==0.0||rtk->ph[rcv-1][sat-1][f]==0.0) {
            continue;
        }
        tt=timediff(obs[i].time,rtk->pt[rcv-1][sat-1][f]);
        lam=rtk->lams[sat-1][f];
        if (lam<=0.0||fabs(tt)<DTTOL) continue;
        
        /* cycle slip threshold (cycle) */
        thres=MAXACC*tt*tt/2.0/lam+rtk->opt.err[4]*fabs(tt)*4.0;
        
        /* phase difference and doppler x time (cycle) */
        dph=obs[i].L[f]-rtk->ph[rcv-1][sat-1][f];
        dpt=-obs[i].D[f]*tt;
        
        if (fabs(dph-dpt)<=thres) continue;
        
        rtk->slip[sat-1][f]|=1;
        
        errmsg(rtk,"slip detected (sat=%3d rcv=%d L%d=%.3f %.3f thres=%.3f)\n",
               sat,rcv,f+1,dph,dpt,thres);
        trace(2,"detslp : slip detected (%s sat=%3d rcv=%d L%d=%.3f %.3f thres=%.3f)\n",
              rtk->tstr,sat,rcv,f+1,dph,dpt,thres);
    }
#endif
}
/* temporal update of phase biases -------------------------------------------*/
static void udbias(rtk_t *rtk, double tt, const obsd_t *obs, const int *sat,
                   const int *iu, const int *ir, int ns)
{
    double cp,pr,*bias,offset,lami;
    int i,j,f,reset;
    
    trace(3,"udbias : tt=%.1f ns=%d\n",tt,ns);
    
    for (i=0;i<ns;i++) {
        
        /* detect cycle slip by LLI */
        for (f=0;f<rtk->opt.nf;f++) rtk->slip[sat[i]-1][f]&=0xFC;
        detslp_lli(rtk,obs,iu[i],1);
        detslp_lli(rtk,obs,ir[i],2);
        
        /* detect cycle slip by geometry-free phase jump */
        detslp_gf(rtk,obs,iu[i],ir[i]);
        
        /* detect cycle slip by doppler and phase difference */
        detslp_dop(rtk,obs,iu[i],1);
        detslp_dop(rtk,obs,ir[i],2);
    }
    for (f=0;f<rtk->opt.nf;f++) {
        /* reset phase-bias if instantaneous AR or expire obs outage counter */
        for (i=1;i<=MAXSAT;i++) {
            
            reset=++rtk->outc[i-1][f]>(unsigned int)rtk->opt.maxout;
            
            if (rtk->opt.modear==2&&rtk->x[IB(i,f,&rtk->opt)]!=0.0) {
                initx(rtk,0.0,0.0,IB(i,f,&rtk->opt));
            }
            else if (reset&&rtk->x[IB(i,f,&rtk->opt)]!=0.0) {
                initx(rtk,0.0,0.0,IB(i,f,&rtk->opt));
                trace(3,"udbias : obs outage counter overflow (%s sat=%3d L%d n=%d)\n",
                      rtk->tstr,i,f+1,rtk->outc[i-1][f]);
            }
            if (rtk->opt.modear!=2&&reset) {
                rtk->lock[i-1][f]=-rtk->opt.minlock;
            }
        }
        /* reset phase-bias if detecting cycle slip */
        for (i=0;i<ns;i++) {
            j=IB(sat[i],f,&rtk->opt);
            rtk->P[j+j*rtk->nx]+=rtk->opt.prn[0]*rtk->opt.prn[0]*tt;
            if (rtk->opt.modear==2||!(rtk->slip[sat[i]-1][f]&1)) continue;
            rtk->x[j]=0.0;
            rtk->lock[sat[i]-1][f]=-rtk->opt.minlock;
        }
        bias=zeros(ns,1);
        
        /* estimate approximate phase-bias by phase - code */
        for (i=j=0,offset=0.0;i<ns;i++) {
            cp=sdobs(obs,iu[i],ir[i],f); /* cycle */
            pr=sdobs(obs,iu[i],ir[i],f+NFREQ);
            lami=rtk->lams[sat[i]-1][f];
            if (cp==0.0||pr==0.0||lami<=0.0) continue;
            
            bias[i]=cp-pr/lami;
            
            if (rtk->x[IB(sat[i],f,&rtk->opt)]!=0.0) {
                offset+=bias[i]-rtk->x[IB(sat[i],f,&rtk->opt)];
                j++;
            }
        }
        /* correct phase-bias offset to enssure phase-code coherency */
        if (j>0) {
            for (i=1;i<=MAXSAT;i++) {
                if (rtk->x[IB(i,f,&rtk->opt)]!=0.0) rtk->x[IB(i,f,&rtk->opt)]+=offset/j;
            }
        }
        /* set initial states of phase-bias */
        for (i=0;i<ns;i++) {
            if (bias[i]==0.0||rtk->x[IB(sat[i],f,&rtk->opt)]!=0.0) continue;
            initx(rtk,bias[i],SQR(rtk->opt.std[0]),IB(sat[i],f,&rtk->opt));
        }
        free(bias);
    }
}
/* temporal update of states --------------------------------------------------*/
static void udstate(rtk_t *rtk, const obsd_t *obs, const int *sat,
                    const int *iu, const int *ir, int ns)
{
    double tt=fabs(rtk->tt),bl,dr[3];
    
    trace(3,"udstate: ns=%d\n",ns);
    
    /* temporal update of position/velocity/acceleration */
    udpos(rtk,tt);
    
    /* temporal update of ionospheric parameters */
    if (rtk->opt.estiono) {
        bl=baseline(rtk->x,rtk->rr[1],dr);
        udion(rtk,tt,bl,sat,ns);
    }
    /* temporal update of tropospheric parameters */
    if (rtk->opt.esttrop) {
        udtrop(rtk,tt);
    }
    /* temporal update of eceiver h/w bias */
    if (rtk->opt.glomodear==2&&(rtk->opt.navsys&SYS_GLO)) {
        udrcvbias(rtk,tt);
    }
    /* temporal update of phase-bias */
    if (rtk->opt.mode>PMODE_DGPS) {
        udbias(rtk,tt,obs,sat,iu,ir,ns);
    }
}
/* undifferenced phase/code residuals ----------------------------------------*/
static int zdres(const obsd_t *obs, int n, const double *rs, const double *dts,
                 double lams[][NFREQ], const double *rr, const prcopt_t *opt,
                 const pcv_t *pcv, double *y, double *e, double *azel)
{
    double r,pos[3],dr[NFREQ]={0},wavelen;
    int i,j;
    
    trace(3,"zdres  : n=%d\n",n);
    
    for (i=0;i<n*opt->nf*2;i++) y[i]=0.0;
    
    if (norm(rr,3)<=0.0) return 0; /* no receiver position */
    
    ecef2pos(rr,pos);
    
    for (i=0;i<n;i++) {
        /* compute geometric-range and azimuth/elevation angle */
        if ((r=geodist(rs+i*3,rr,e+i*3))<=0.0) continue;
        if (satazel(pos,e+i*3,azel+i*2)<opt->elmin) continue;
        
        /* satellite clock-bias and troposphere delay model */
        r+=-CLIGHT*dts[i]+tropmodel(pos,azel+i*2,0.7);
        
        /* antenna phase center correction */
        if (pcv) antmodel(pcv,azel+i*2,dr);
        
        for (j=0;j<opt->nf;j++) {
            
            if ((wavelen=lams[obs[i].sat-1][j])<=0.0) continue;
            
            /* check snr mask */
            if (opt->snrmin>0) {
                if (obs[i].SNR[j]<opt->snrmin) continue;
            }
            /* residuals = observable - pseudorange */
            if (obs[i].L[j]!=0.0) y[j+i*opt->nf*2]=wavelen*obs[i].L[j]-r-dr[j];
            if (obs[i].P[j]!=0.0) y[j+i*opt->nf*2+opt->nf]=obs[i].P[j]-r-dr[j];
        }
    }
    trace(4,"rr =%.3f %.3f %.3f\n",rr[0],rr[1],rr[2]);
    trace(4,"pos=%.9f %.9f %.3f\n",pos[0]*R2D,pos[1]*R2D,pos[2]);
    for (i=0;i<n;i++) {
        trace(4,"sat=%3d %13.3f %13.3f %13.3f %13.10f %6.1f %5.1f\n",
              obs[i].sat,rs[i*3],rs[1+i*3],rs[2+i*3],dts[i],azel[i*2]*R2D,
              azel[1+i*2]*R2D);
    }
    trace(4,"y=\n"); tracemat(4,y,opt->nf*2,n,13,3);
    
    return 1;
}
/* test valid observation data -----------------------------------------------*/
static int validobs(int i, int j, int f, int nf, double *y)
{
    /* if no phase observable, psudorange is also unusable */
    return y[f+i*nf*2]!=0.0&&y[f+j*nf*2]!=0.0&&
           (f<nf||(y[f-nf+i*nf*2]!=0.0&&y[f-nf+j*nf*2]!=0.0));
}
/* double-differenced measurement error covariance ---------------------------*/
static void ddcov(const int *nb, int n, const double *Ri, const double *Rj,
                  int nv, double *R)
{
    int i,j,k=0,b;
    
    trace(3,"ddcov  : n=%d\n",n);
    
    for (i=0;i<nv*nv;i++) R[i]=0.0;
    for (b=0;b<n;k+=nb[b++]) {
        
        for (i=0;i<nb[b];i++) for (j=0;j<nb[b];j++) {
            R[k+i+(k+j)*nv]=Ri[k+i]+(i==j?Rj[k+i]:0.0);
        }
    }
    trace(5,"R=\n"); tracemat(5,R,nv,nv,8,6);
}
/* baseline length constraint ------------------------------------------------*/
static int constbl(rtk_t *rtk, const double *x, const double *P, double *v,
                   double *H, double *Ri, double *Rj, int index)
{
    const double thres=0.1; /* threshold for nonliearity (v.2.3.0) */
    double tt,xb[3],b[3],bb,var=0.0;
    int i;
     
    trace(3,"constbl: \n");
    
    /* no constraint */
    if (rtk->opt.baseline[0]<=0.0) return 0;
    
    /* time-adjusted baseline vector and length */
    tt=timediff(rtk->time[0],rtk->time[1]);
    for (i=0;i<3;i++) {
        xb[i]=rtk->rr[1][i]+rtk->rr[1][i+3]*tt;
        b[i]=x[i]-xb[i];
    }
    bb=norm(b,3);
    
    /* approximate variance of solution */
    if (P) {
        for (i=0;i<3;i++) var+=P[i+i*rtk->nx];
        var/=3.0;
    }
    /* check nonlinearity */
    if (var>thres*thres*bb*bb) {
        trace(3,"constbl : equation nonlinear (bb=%.3f var=%.3f)\n",bb,var);
        return 0;
    }
    /* constraint to baseline length */
    v[index]=rtk->opt.baseline[0]-bb;
    if (H) {
        for (i=0;i<3;i++) H[i+index*rtk->nx]=b[i]/bb;
    }
    Ri[index]=0.0;
    Rj[index]=SQR(rtk->opt.baseline[1]);
    
    trace(4,"baseline len   v=%13.3f R=%8.6f %8.6f\n",v[index],Ri[index],Rj[index]);
    
    return 1;
}
/* double-differenced phase/code residuals -----------------------------------*/
static int ddres(rtk_t *rtk, double dt, const double *x, const double *P,
                 const int *sat, double *y, double *e, double *azel,
                 const int *iu, const int *ir, int ns, double *v, double *H,
                 double *R, int *vflg)
{
    prcopt_t *opt=&rtk->opt;
    double bl,dr[3],posu[3],posr[3],didxi=0.0,didxj=0.0;
    double *im,*tmwu,*tmwr,*Ri,*Rj,s,lami,lamj,fi,fj,df,*Hi=NULL;
    int i,j,k,m,f,ff,nv=0,nb[NFREQ*2*2+1]={0},b=0,sysi,sysj;
    
    trace(3,"ddres  : dt=%.1f nx=%d ns=%d\n",dt,rtk->nx,ns);
    
    bl=baseline(x,rtk->rr[1],dr);
    ecef2pos(x,posu); ecef2pos(rtk->rr[1],posr);
    
    Ri=mat(ns*opt->nf*2+1,1); Rj=mat(ns*opt->nf*2+1,1);
    im=mat(ns,1); tmwu=mat(ns,1); tmwr=mat(ns,1);
    
    for (i=0;i<MAXSAT;i++) for (j=0;j<NFREQ;j++) {
        rtk->resp[i][j]=rtk->resc[i][j]=0.0;
    }
    /* compute factors of ionospheric and tropospheric delay */
    for (i=0;i<ns;i++) {
        if (opt->estiono) {
            im[i]=(ionmapf(posu,azel+iu[i]*2)+ionmapf(posr,azel+ir[i]*2))/2.0;
        }
        if (opt->esttrop) {
            tropmapf(rtk->time[0],posu,azel+iu[i]*2,tmwu+i);
            tropmapf(rtk->time[0],posr,azel+ir[i]*2,tmwr+i);
        }
    }
    for (m=0;m<2;m++) /* for each system (0:gps,1:glonass) */
    
    for (f=opt->mode>PMODE_DGPS?0:opt->nf;f<opt->nf*2;f++) {
        
        /* search reference satellite with highest elevation */
        for (i=-1,j=0;j<ns;j++) {
            sysi=rtk->sys[sat[j]-1];
            if ((m==0&&sysi==SYS_GLO)||(m==1&&sysi!=SYS_GLO)) continue;
            if (!validobs(iu[j],ir[j],f,opt->nf,y)) continue;
            if (i<0||azel[1+iu[j]*2]>=azel[1+iu[i]*2]) i=j;
        }
        if (i<0) continue;
        
        /* make double difference */
        for (j=0;j<ns;j++) {
            if (i==j) continue;
            sysi=rtk->sys[sat[i]-1];
            sysj=rtk->sys[sat[j]-1];
            if ((m==0&&sysj==SYS_GLO)||(m==1&&sysj!=SYS_GLO)) continue;
            if (!validobs(iu[j],ir[j],f,opt->nf,y)) continue;
            
            ff=f%opt->nf;
            lami=rtk->lams[sat[i]-1][ff];
            lamj=rtk->lams[sat[j]-1][ff];
            if (lami<=0.0||lamj<=0.0) continue;
            if (H) Hi=H+nv*rtk->nx;
            
            /* double-differenced residual */
            v[nv]=(y[f+iu[i]*opt->nf*2]-y[f+ir[i]*opt->nf*2])-
                  (y[f+iu[j]*opt->nf*2]-y[f+ir[j]*opt->nf*2]);
            
            /* partial derivatives by rover position */
            if (H) {
                for (k=0;k<3;k++) {
                    Hi[k]=-e[k+iu[i]*3]+e[k+iu[j]*3];
                }
            }
            /* double-differenced ionospheric delay term */
            if (opt->estiono) {
                fi=lami/lam[0]; fj=lamj/lam[0];
                didxi=(f<opt->nf?-1.0:1.0)*fi*fi*im[i];
                didxj=(f<opt->nf?-1.0:1.0)*fj*fj*im[j];
                v[nv]-=didxi*x[II(sat[i],opt)]-didxj*x[II(sat[j],opt)];
                if (H) {
                    Hi[II(sat[i],opt)]= didxi;
                    Hi[II(sat[j],opt)]=-didxj;
                }
            }
            /* double-differenced tropospheric delay term */
            if (opt->esttrop) {
                v[nv]-=(tmwu[i]-tmwu[j])*x[IT(0,opt)]-(tmwr[i]-tmwr[j])*x[IT(1,opt)];
                if (H) {
                    Hi[IT(0,opt)]= (tmwu[i]-tmwu[j]);
                    Hi[IT(1,opt)]=-(tmwr[i]-tmwr[j]);
                }
            }
            /* double-differenced phase-bias term */
            if (f<opt->nf) {
                v[nv]-=lami*x[IB(sat[i],f,opt)]-lamj*x[IB(sat[j],f,opt)];
                if (H) {
                    Hi[IB(sat[i],f,opt)]= lami;
                    Hi[IB(sat[j],f,opt)]=-lamj;
                }
            }
            /* glonass receiver h/w bias term */
            if (rtk->opt.glomodear==2&&sysi==SYS_GLO&&sysj==SYS_GLO&&ff<NFREQGLO) {
                df=(CLIGHT/lami-CLIGHT/lamj)/1E6; /* freq-difference (MHz) */
                v[nv]-=df*x[IL(ff,opt)];
                if (H) Hi[IL(ff,opt)]=df;
            }
            /* test innovation */
            if (opt->maxinno>0.0&&fabs(v[nv])>opt->maxinno) {
                if (f<opt->nf) {
                    rtk->rejc[sat[i]-1][f]++;
                    rtk->rejc[sat[j]-1][f]++;
                }
                errmsg(rtk,"outlier rejected (sat=%3d-%3d %s%d v=%.3f)\n",
                       sat[i],sat[j],f<opt->nf?"L":"P",f%opt->nf+1,v[nv]);
                trace(1,"ddres  : outlier rejected (%s sat=%3d-%3d %s%d v=%.3f)\n",
                      rtk->tstr,sat[i],sat[j],f<opt->nf?"L":"P",f%opt->nf+1,v[nv]);
                continue;
            }
            /* single-differenced measurement error variances */
            Ri[nv]=varerr(sysi,azel[1+iu[i]*2],bl,dt,f,opt);
            Rj[nv]=varerr(sysj,azel[1+iu[j]*2],bl,dt,f,opt);
            
            /* set valid data flags */
            if (opt->mode>PMODE_DGPS) {
                if (f<opt->nf) rtk->vsat[sat[i]-1][f]=rtk->vsat[sat[j]-1][f]=1;
            }
            else {
                rtk->vsat[sat[i]-1][f-opt->nf]=rtk->vsat[sat[j]-1][f-opt->nf]=1;
            }
            if (f<opt->nf) {
                rtk->resc[sat[j]-1][f]=v[nv];
            }
            else {
                rtk->resp[sat[j]-1][f-opt->nf]=v[nv];
            }
            trace(4,"sat=%3d-%3d %s%d v=%13.3f R=%8.6f %8.6f\n",sat[i],
                  sat[j],f<opt->nf?"L":"P",f%opt->nf+1,v[nv],Ri[nv],Rj[nv]);
            
            vflg[nv++]=(sat[i]<<16)|(sat[j]<<8)|((f<opt->nf?0:1)<<4)|(f%opt->nf);
            nb[b]++;
        }
        /* restore single-differenced residuals assuming sum equal zero */
        if (f<opt->nf) {
            for (j=0,s=0.0;j<MAXSAT;j++) s+=rtk->resc[j][f];
            s/=nb[b]+1;
            for (j=0;j<MAXSAT;j++) {
                if (j==sat[i]-1||rtk->resc[j][f]!=0.0) rtk->resc[j][f]-=s;
            }
        }
        else {
            for (j=0,s=0.0;j<MAXSAT;j++) s+=rtk->resp[j][f-opt->nf];
            s/=nb[b]+1;
            for (j=0;j<MAXSAT;j++) {
                if (j==sat[i]-1||rtk->resp[j][f-opt->nf]!=0.0) rtk->resp[j][f-opt->nf]-=s;
            }
        }
        b++;
    }
    /* end of system loop */
    
    /* baseline length constraint for moving baseline */
    if (opt->mode==PMODE_MOVEB&&constbl(rtk,x,P,v,H,Ri,Rj,nv)) {
        vflg[nv++]=3<<4;
        nb[b++]++;
    }
    if (H) {trace(5,"H=\n"); tracemat(5,H,rtk->nx,nv,7,4);}
    
    /* double-differenced measurement error covariance */
    ddcov(nb,b,Ri,Rj,nv,R);
    
    free(Ri); free(Rj); free(im); free(tmwu); free(tmwr);
    
    return nv;
}
/* time-interpolation of residuals (for post-mission) ------------------------*/
static double intpres(gtime_t time, const obsd_t *obs, int n, const nav_t *nav,
                      double lams[][NFREQ], const double *rb,
                      const prcopt_t *opt, double *y)
{
    static obsd_t obsb[MAXOBS];
    static double yb[MAXOBS*NFREQ*2];
    static double rs[MAXOBS*3],dts[MAXOBS],e[MAXOBS*3],azel[MAXOBS*2];
    static int nb=0;
    double tt=timediff(time,obs[0].time),ttb,*p,*q;
    int i,j,k;
    
    trace(3,"intpres: n=%d tt=%.1f\n",n,tt);
    
    if (nb==0||fabs(tt)<DTTOL) {
        nb=n; for (i=0;i<n;i++) obsb[i]=obs[i];
        return tt;
    }
    ttb=timediff(time,obsb[0].time);
    if (fabs(ttb)>opt->maxtdiff*2.0||ttb==tt) return tt;
    
    if (opt->sateph) {
        satposp(time,obsb,nb,nav,rs,dts);
    }
    else {
        satpos(time,obsb,nb,nav,rs,dts);
    }
    if (!zdres(obsb,nb,rs,dts,lams,rb,opt,opt->pcvr,yb,e,azel)) return tt;
    
    for (i=0;i<n;i++) {
        for (j=0;j<nb;j++) if (obsb[j].sat==obs[i].sat) break;
        if (j>=nb) continue;
        for (k=0,p=y+i*opt->nf*2,q=yb+j*opt->nf*2;k<opt->nf*2;k++,p++,q++) {
            if (*p==0.0||*q==0.0) *p=0.0; else *p=(ttb*(*p)-tt*(*q))/(ttb-tt);
        }
    }
    return fabs(ttb)>fabs(tt)?ttb:tt;
}
/* single to double-difference transformation matrix (D') --------------------*/
static int ddmat(rtk_t *rtk, double *D)
{
    int i,j,k,m,f,nb=0,nx=rtk->nx,na=rtk->na;
    
    trace(3,"ddmat :\n");
    
    for (i=0;i<MAXSAT;i++) for (j=0;j<NFREQ;j++) {
        rtk->fix[i][j]=0;
    }
    for (i=0;i<na;i++) D[i+i*nx]=1.0;
    
    for (m=0;m<2;m++) { /* 0:gps,1:glonass */
        
        if (m==1&&rtk->opt.glomodear==0) continue;
        
        for (f=0,k=na;f<rtk->opt.nf;f++,k+=MAXSAT) {
            
            for (i=k;i<k+MAXSAT;i++) {
                if (rtk->x[i]==0.0||
                    (m==0&&rtk->sys[i-k]==SYS_GLO)||
                    (m==1&&rtk->sys[i-k]!=SYS_GLO)) continue;
                
                if (rtk->lock[i-k][f]>0&&!(rtk->slip[i-k][f]&2)) {
                    rtk->fix[i-k][f]=2;
                    break;
                }
                else rtk->fix[i-k][f]=1;
            }
            for (j=k;j<k+MAXSAT;j++) {
                if (i==j||rtk->x[j]==0.0||
                    (m==0&&rtk->sys[j-k]==SYS_GLO)||
                    (m==1&&rtk->sys[j-k]!=SYS_GLO)) continue;
                
                if (rtk->lock[j-k][f]>0&&!(rtk->slip[j-k][f]&2)) {
                    D[i+(na+nb)*nx]= 1.0;
                    D[j+(na+nb)*nx]=-1.0;
                    nb++;
                    rtk->fix[j-k][f]=2;
                }
                else rtk->fix[j-k][f]=1;
            }
        }
    }
    trace(5,"D=\n"); tracemat(5,D,nx,na+nb,2,0);
    return nb;
}
/* restore single-differenced ambiguity --------------------------------------*/
static void restamb(rtk_t *rtk, const double *bias, int nb, double *xa)
{
    int i,n,m,f,index[MAXSAT],nv=0;
    
    for (i=0;i<rtk->nx;i++) xa[i]=rtk->x [i];
    for (i=0;i<rtk->na;i++) xa[i]=rtk->xa[i];
    
    for (m=0;m<2;m++) for (f=0;f<rtk->opt.nf;f++) {
        
        for (n=i=0;i<MAXSAT;i++) {
            if ((m==0&&rtk->sys[i]==SYS_GLO)||
                (m==1&&rtk->sys[i]!=SYS_GLO)||rtk->fix[i][f]!=2) continue;
            index[n++]=IB(i+1,f,&rtk->opt);
        }
        if (n<2) continue;
        
        xa[index[0]]=rtk->x[index[0]];
        
        for (i=1;i<n;i++) {
            xa[index[i]]=xa[index[0]]-bias[nv++];
        }
    }
}
/* hold integer ambiguity ----------------------------------------------------*/
static void holdamb(rtk_t *rtk, const double *bias, int nb)
{
    double *v,*H,*R;
    int i,n,m,f,info,index[MAXSAT],nv=0;
    
    v=mat(nb,1); H=zeros(nb,rtk->nx); R=zeros(nb,nb);
    
    for (m=0;m<2;m++) for (f=0;f<rtk->opt.nf;f++) {
        
        for (n=i=0;i<MAXSAT;i++) {
            if ((m==0&&rtk->sys[i]==SYS_GLO)||
                (m==1&&rtk->sys[i]!=SYS_GLO)||rtk->fix[i][f]!=2) continue;
            index[n++]=IB(i+1,f,&rtk->opt);
            rtk->fix[i][f]=3; /* hold */
        }
        /* constraint to fixed ambiguity */
        for (i=1;i<n;i++) {
            v[nv]=bias[nv]-(rtk->x[index[0]]-rtk->x[index[i]]);
            
            H[index[0]+nv*rtk->nx]= 1.0;
            H[index[i]+nv*rtk->nx]=-1.0;
            
            R[nv+nv*nb]=VAR_FIXAMB;
            nv++;
        }
    }
    /* update states with constraints */
    if ((info=filter(rtk->x,rtk->P,H,v,R,rtk->nx,nv))) {
        errmsg(rtk,"filter error (info=%d)\n",info);
        trace(1,"holdamb: filter error (%s info=%d)\n",rtk->tstr,info);
    }
    free(v); free(H); free(R);
}
/* resolve integer ambiguity -------------------------------------------------*/
static int resamb(rtk_t *rtk, double *bias, double *xa)
{
    prcopt_t *opt=&rtk->opt;
    int i,j,ny,nb,info,nx=rtk->nx,na=rtk->na;
    double *D,*DP,*y,*Qy,*b,*db,*Qb,*Qab,*QQ,s[2];
    
    trace(3,"resamb : nx=%d\n",nx);
    
    rtk->ambratio=0.0;
    
    if (rtk->opt.mode<=PMODE_DGPS||rtk->opt.modear<=0||rtk->opt.thresar<1.0) {
        return 0;
    }
    /* single to double-difference transformation matrix (D') */
    D=zeros(nx,nx);
    if ((nb=ddmat(rtk,D))<=0) {
        errmsg(rtk,"no valid double-difference\n");
        trace(1,"resamb : no valid double-difference (%s)\n",rtk->tstr);
        free(D);
        return 0;
    }
    ny=na+nb; y=mat(ny,1); Qy=mat(ny,ny); DP=mat(ny,nx);
    b=mat(nb,2); db=mat(nb,1); Qb=mat(nb,nb); Qab=mat(na,nb); QQ=mat(na,nb);
    
    /* transform single to double-differenced phase-bias (y=D'*x, Qy=D'*P*D) */
    matmul("TN",ny, 1,nx,1.0,D ,rtk->x,0.0,y );
    matmul("TN",ny,nx,nx,1.0,D ,rtk->P,0.0,DP);
    matmul("NN",ny,ny,nx,1.0,DP,D     ,0.0,Qy);
    
    /* phase-bias covariance (Qb) and real-parameters to bias covariance (Qab) */
    for (i=0;i<nb;i++) for (j=0;j<nb;j++) Qb [i+j*nb]=Qy[na+i+(na+j)*ny];
    for (i=0;i<na;i++) for (j=0;j<nb;j++) Qab[i+j*na]=Qy[   i+(na+j)*ny];
    
    trace(4,"N(0)="); tracemat(4,y+na,1,nb,10,3);
    
    /* lambda/mlambda integer least-square estimation */
    if (!(info=lambda(nb,2,y+na,Qb,b,s))) {
        
        trace(4,"N(1)="); tracemat(4,b   ,1,nb,10,3);
        trace(4,"N(2)="); tracemat(4,b+nb,1,nb,10,3);
        
        rtk->ambratio=s[0]>0?s[1]/s[0]:0.0;
        if (rtk->ambratio>999.9) rtk->ambratio=999.9;
        
        /* validation by popular ratio-test */
        if (s[0]<=0.0||s[1]/s[0]>=opt->thresar) {
            
            /* transform float to fixed solution (xa=xa-Qab*Qb\(b0-b)) */
            for (i=0;i<na;i++) {
                rtk->xa[i]=rtk->x[i];
                for (j=0;j<na;j++) rtk->Pa[i+j*na]=rtk->P[i+j*nx];
            }
            for (i=0;i<nb;i++) {
                bias[i]=b[i];
                y[na+i]-=b[i];
            }
            if (!matinv(Qb,nb)) {
                matmul("NN",nb,1,nb, 1.0,Qb ,y+na,0.0,db);
                matmul("NN",na,1,nb,-1.0,Qab,db  ,1.0,rtk->xa);
                
                /* covariance of fixed solution (Qa=Qa-Qab*Qb^-1*Qab') */
                matmul("NN",na,nb,nb, 1.0,Qab,Qb ,0.0,QQ);
                matmul("NT",na,na,nb,-1.0,QQ ,Qab,1.0,rtk->Pa);
                
                trace(3,"resamb : validation ok (%s nb=%d ratio=%.2f s=%.2f/%.2f)\n",
                      rtk->tstr,nb,s[0]==0.0?0.0:s[1]/s[0],s[0],s[1]);
                
                /* restore single-differenced ambiguity */
                restamb(rtk,bias,nb,xa);
            }
            else nb=0;
        }
        else { /* validation failed */
            errmsg(rtk,"ambiguity validation failed (nb=%d ratio=%.2f s=%.2f/%.2f)\n",
                   nb,s[1]/s[0],s[0],s[1]);
            trace(2,"resamb : validation failed (%s nb=%d ratio=%.2f s=%.2f/%.2f)\n",
                  rtk->tstr,nb,s[1]/s[0],s[0],s[1]);
            nb=0;
        }
    }
    else {
        errmsg(rtk,"lambda error (info=%d)\n",info);
        trace(1,"resamb : lambda error (%s info=%d)\n",rtk->tstr,info);
    }
    free(D); free(y); free(Qy); free(DP);
    free(b); free(db); free(Qb); free(Qab); free(QQ);
    
    return nb; /* number of ambiguities */
}
/* validation of solution ----------------------------------------------------*/
static int valpos(rtk_t *rtk, const double *v, const double *R, const int *vflg,
                  int nv, double thres)
{
    prcopt_t *opt=&rtk->opt;
    double vv=0.0,fact=thres*thres;
    int i,stat=1,sat1,sat2,type,freq;
    char *stype;
    
    trace(3,"valpos : nv=%d thres=%.1f\n",nv,thres);
    
    /* post-fit residual test */
    for (i=0;i<nv;i++) {
        if (v[i]*v[i]<=fact*R[i+i*nv]) continue;
        sat1=(vflg[i]>>16)&0xFF;
        sat2=(vflg[i]>> 8)&0xFF;
        type=(vflg[i]>> 4)&0xF;
        freq=vflg[i]&0xF;
        stype=type==0?"L":(type==1?"L":"C");
        errmsg(rtk,"large residual (sat=%d-%d %s%d v=%6.3f sig=%.3f)\n",
              sat1,sat2,stype,freq+1,v[i],SQRT(R[i+i*nv]));
        trace(2,"valpos : large residual (%s sat=%d-%d %s%d v=%6.3f sig=%.3f)\n",
              rtk->tstr,sat1,sat2,stype,freq+1,v[i],SQRT(R[i+i*nv]));
    }
    /* chi-square validation */
    if (stat&&nv>NP(opt)) {
        
        for (i=0;i<nv;i++) vv+=v[i]*v[i]/R[i+i*nv];
        
        if (vv>chisqr[nv-NP(opt)-1]) {
            errmsg(rtk,"residuals validation failed (nv=%d np=%d vv=%.2f cs=%.2f)\n",
                   nv,NP(opt),vv,chisqr[nv-NP(opt)-1]);
            trace(2,"valpos : validation failed (%s nv=%d np=%d vv=%.2f cs=%.2f)\n",
                  rtk->tstr,nv,NP(opt),vv,chisqr[nv-NP(opt)-1]);
            stat=0;
        }
        else {
            trace(3,"valpos : validation ok (%s nv=%d np=%d vv=%.2f cs=%.2f)\n",
                  rtk->tstr,nv,NP(opt),vv,chisqr[nv-NP(opt)-1]);
        }
    }
    return stat;
}
/* relative positioning ------------------------------------------------------*/
static int relpos(rtk_t *rtk, const obsd_t *obs, int nu, int nr,
                  const nav_t *nav)
{
    prcopt_t *opt=&rtk->opt;
    gtime_t time=obs[0].time;
    double *rs,*dts,*y,*e,*azel,*v,*H,*R,*xp,*Pp,*xa,*bias,dt;
    int i,j,f,n=nu+nr,ns,ny,nv,nb,sat[MAXSAT],iu[MAXSAT],ir[MAXSAT],niter;
    int info,stat,vflg[MAXOBS*NFREQ*2+1];
    
    trace(3,"relpos : nx=%d nu=%d nr=%d\n",rtk->nx,nu,nr);
    
    stat=rtk->opt.mode<=PMODE_DGPS?SOLQ_DGPS:SOLQ_FLOAT;
    
    dt=timediff(time,obs[nu].time);
    
    rs=mat(3,n); dts=mat(1,n); y=mat(opt->nf*2,n); e=mat(3,n); azel=zeros(2,n);
    bias=mat(rtk->nx,1);
    
    for (i=0;i<MAXSAT;i++) {
        rtk->sys[i]=satsys(i+1,NULL);
        for (j=0;j<NFREQ;j++) rtk->vsat[i][j]=rtk->snr[i][j]=0;
    }
    /* compute satellite position/clock-bias */
    if (rtk->opt.sateph) {
        satposp(time,obs,n,nav,rs,dts); /* by precise ephemeris */
    }
    else {
        satpos(time,obs,n,nav,rs,dts); /* by broadcast ephemeris */
    }
    /* satellite carrier wave length */
    for (i=0;i<n;i++) for (j=0;j<rtk->opt.nf;j++) {
        if (!(rtk->sys[obs[i].sat-1]&rtk->opt.navsys)) continue;
        rtk->lams[obs[i].sat-1][j]=satwavelen(obs[i].time,obs[i].sat,j,nav);
    }
    /* undifferenced residuals for base station */
    if (!zdres(obs+nu,nr,rs+nu*3,dts+nu,rtk->lams,rtk->rr[1],opt,opt->pcvr,
               y+nu*opt->nf*2,e+nu*3,azel+nu*2)) {
        errmsg(rtk,"initial base station position error\n");
        trace(2,"relpos : initial base station postion error (%s)\n",rtk->tstr);
        free(rs); free(dts); free(y); free(e); free(azel);
        return SOLQ_NONE;
    }
    /* time-interpolation of residuals (for post-processing) */
    if (opt->intpref) {
        dt=intpres(time,obs+nu,nr,nav,rtk->lams,rtk->rr[1],opt,y+nu*opt->nf*2);
    }
    /* select common satellites between rover and base-station */
    if ((ns=selsat(obs,azel,nu,nr,opt,sat,iu,ir))<=0) {
        errmsg(rtk,"no common satellite\n");
        trace(1,"relpos : no common satellite (%s)\n",rtk->tstr);
        free(rs); free(dts); free(y); free(e); free(azel);
        return SOLQ_NONE;
    }
    /* temporal update of states */
    udstate(rtk,obs,sat,iu,ir,ns);
    
    trace(4,"x(0)="); tracemat(4,rtk->x,1,NR(opt),13,4);
    
    xp=mat(rtk->nx,1); Pp=zeros(rtk->nx,rtk->nx); xa=mat(rtk->nx,1);
    matcpy(xp,rtk->x,rtk->nx,1);
    
    ny=ns*opt->nf*2+1;
    v=mat(ny,1); H=zeros(rtk->nx,ny); R=mat(ny,ny);
    
    /* add 2 iterations for baseline-constraint moving-base */
    niter=opt->niter+(opt->mode==PMODE_MOVEB&&opt->baseline[0]>0.0?2:0);
    
    for (i=0;i<niter;i++) {
        /* undifferenced residuals for rover */
        if (!zdres(obs,nu,rs,dts,rtk->lams,xp,opt,opt->pcvu,y,e,azel)) {
            trace(1,"relpos : rover initial position error (%s)\n",rtk->tstr);
            stat=SOLQ_NONE;
            break;
        }
        /* double-differenced residuals and partial derivatives */
        if ((nv=ddres(rtk,dt,xp,Pp,sat,y,e,azel,iu,ir,ns,v,H,R,vflg))<1) {
            errmsg(rtk,"no double-differenced residual\n");
            trace(1,"relpos : no double-differenced residual (%s)\n",rtk->tstr);
            stat=SOLQ_NONE;
            break;
        }
        /* kalman filter measurement update */
        matcpy(Pp,rtk->P,rtk->nx,rtk->nx);
        if ((info=filter(xp,Pp,H,v,R,rtk->nx,nv))) {
            errmsg(rtk,"filter error (info=%d)\n",info);
            trace(1,"relpos : filter error (%s info=%d)\n",rtk->tstr,info);
            stat=SOLQ_NONE;
            break;
        }
        trace(4,"x(%d)=",i+1); tracemat(4,xp,1,NR(opt),13,4);
    }
    if (stat!=SOLQ_NONE&&
        zdres(obs,nu,rs,dts,rtk->lams,xp,opt,opt->pcvu,y,e,azel)) {
        
        /* post-fit residuals for float solution */
        nv=ddres(rtk,dt,xp,Pp,sat,y,e,azel,iu,ir,ns,v,NULL,R,vflg);
        
        /* validation of float solution */
        if (valpos(rtk,v,R,vflg,nv,4.0)) {
            
            /* update state and covariance matrix */
            matcpy(rtk->x,xp,rtk->nx,1);
            matcpy(rtk->P,Pp,rtk->nx,rtk->nx);
            
            /* update ambiguity control struct */
            rtk->ns=0;
            for (i=0;i<ns;i++) for (f=0;f<opt->nf;f++) {
                if (!rtk->vsat[sat[i]-1][f]) continue;
                rtk->lock[sat[i]-1][f]++;
                rtk->outc[sat[i]-1][f]=0;
                if (azel[1+iu[i]*2]<opt->elmaskar) rtk->lock[sat[i]-1][f]=0;
                if (f==0) rtk->ns++; /* valid satellite count by L1 */
            }
            /* lack of valid satellites */
            if (rtk->ns<4) stat=SOLQ_NONE;
        }
        else stat=SOLQ_NONE;
    }
    /* resolve integer ambiguity */
    if (stat!=SOLQ_NONE&&(nb=resamb(rtk,bias,xa))>1) {
        
        if (zdres(obs,nu,rs,dts,rtk->lams,xa,opt,opt->pcvu,y,e,azel)) {
            
            /* post-fit reisiduals for fixed solution */
            nv=ddres(rtk,dt,xa,NULL,sat,y,e,azel,iu,ir,ns,v,NULL,R,vflg);
            
            /* validation of fixed solution */
            if (valpos(rtk,v,R,vflg,nv,4.0)) {
                
                /* hold integer ambiguity */
                if (++rtk->nfix>=NFIXTOHOLD&&rtk->opt.modear==3) {
                    holdamb(rtk,bias,nb);
                }
                stat=SOLQ_FIX;
            }
        }
    }
    else rtk->nfix=0;
    
    /* save solution status */
    for (i=0;i<n;i++) for (j=0;j<rtk->opt.nf;j++) {
        if (obs[i].L[j]==0.0) continue;
        rtk->pt[obs[i].rcv-1][obs[i].sat-1][j]=obs[i].time;
        rtk->ph[obs[i].rcv-1][obs[i].sat-1][j]=obs[i].L[j];
    }
    for (i=0;i<ns;i++) for (j=0;j<rtk->opt.nf;j++) {
        rtk->snr[sat[i]-1][j]=MIN(obs[iu[i]].SNR[j],obs[ir[i]].SNR[j]);
    }
    for (i=0;i<MAXSAT;i++) for (j=0;j<rtk->opt.nf;j++) {
        if (rtk->fix[i][j]==2&&stat!=SOLQ_FIX) rtk->fix[i][j]=1;
        if (rtk->slip[i][j]&1) rtk->slipc[i][j]++;
    }
    free(rs); free(dts); free(y); free(e); free(azel);
    free(xp); free(Pp); free(xa); free(v); free(H); free(R); free(bias);
    
    return stat;
}
/* initialize rtk control ------------------------------------------------------
* initialize rtk control struct
* args   : rtk_t    *rtk    IO  rtk control/result struct
*          prcopt_t *opt    I   positioning options (see rtklib.h)
* return : none
*-----------------------------------------------------------------------------*/
extern void rtkinit(rtk_t *rtk, const prcopt_t *opt)
{
    rtk_t rtk0={{{0}}};
    *rtk=rtk0;
    rtk->nx=NR(opt)+MAXSAT*opt->nf;
    rtk->na=NR(opt);
    rtk->x =zeros(rtk->nx,1);
    rtk->P =zeros(rtk->nx,rtk->nx);
    rtk->xa=zeros(rtk->na,1);
    rtk->Pa=zeros(rtk->na,rtk->na);
    rtk->opt=*opt;
}
/* free rtk control ------------------------------------------------------------
* free memory for rtk control struct
* args   : rtk_t    *rtk    IO  rtk control/result struct
* return : none
*-----------------------------------------------------------------------------*/
extern void rtkfree(rtk_t *rtk)
{
    free(rtk->x ); rtk->x =NULL;
    free(rtk->P ); rtk->P =NULL;
    free(rtk->xa); rtk->xa=NULL;
    free(rtk->Pa); rtk->Pa=NULL;
}
/* estimate receiver position/velocity by single point positining ------------*/
static int rcvposv(rtk_t *rtk, const obsd_t *obs, int nobs, const nav_t *nav,
                   const sbsstat_t *sbsstat, int index, double *Qr, double *Qv)
{
    const char *errs[]={
        "# of sats","# of valid sats","lsq error","iter divergent",
        "validation error","gdop error"
    };
    const char *err="";
    gtime_t time;
    double rr[6]={0},dtr[2],*azel,*res;
    double elmin=rtk->opt.elmin,snrmin=rtk->opt.snrmin;
    int i,corr,vsat[MAXOBS]={0},sat,stat=SOLQ_SINGLE;
    
    trace(3,"rcvposv : nobs=%d neph=%d\n",nobs,nav->n);
    
    azel=mat(nobs,2); res=mat(nobs,1);
    
    /* estimate position with pseudorange */
    if (rtk->opt.mode==PMODE_SINGLE&&sbsstat&&rtk->opt.sbascorr) {
        rtk->ns=sbspntpos(obs,nobs,nav,sbsstat,elmin,snrmin,rtk->opt.sbascorr,
                          rr,Qr,dtr,azel,vsat,&corr);
        if (corr) stat=SOLQ_SBAS;
    }
    else {
        rtk->ns=pntpos(obs,nobs,nav,elmin,snrmin,rr,Qr,dtr,azel,vsat,res);
    }
    /* set azimuth/elevation angles, valid sat flags, residual and snr */
    for (i=0;i<MAXSAT;i++) rtk->vs[index][i]=0;
    for (i=0;i<nobs;i++) {
        sat=obs[i].sat;
        rtk->az[index][sat-1]=azel[i*2];
        rtk->el[index][sat-1]=azel[1+i*2];
        rtk->vs[index][sat-1]=vsat[i];
        if (index==0) {
            rtk->vsat[sat-1][0]=vsat[i];
            rtk->resp[sat-1][0]=res[i];
            rtk->snr [sat-1][0]=obs[i].SNR[0];
        }
    }
    if (rtk->ns<4) {
        if (-6<=rtk->ns&&rtk->ns<=-1) err=errs[-rtk->ns-1];
        errmsg(rtk,"single point positioning error (%s)\n",err);
        trace(1,"rtkpos : single point positioning error (%s %s)\n",rtk->tstr,err);
        
        free(azel); free(res);
        return SOLQ_NONE;
    }
    /* time of receiver clock */
    time=timeadd(obs[0].time,-dtr[0]);
    if (index==0&&rtk->time[0].time!=0) rtk->tt=timediff(time,rtk->time[0]);
    rtk->time[index]=time;
    
    /* estimate velocity with doppler frequency */
    if (pntvel(obs,nobs,nav,rr,azel,vsat,rr+3,Qv,dtr+1)<4) {
        for (i=0;i<3;i++) rr[i+3]=0.0;
    }
    for (i=0;i<6;i++) rtk->rr[index][i]=rr[i];
    
    free(azel); free(res);
    
    return stat;
}
/* rtk positioning -------------------------------------------------------------
* input observation data and navigation message, compute rover position by 
* relative positioning and output results.
* args   : rtk_t *rtk       IO  rtk control/result struct
*            rtk->time[0]   O   signal reception time rover (gpst)
*            rtk->time[1]   O   signal reception time base (gpst)
*            rtk->nx        I   number of all states
*            rtk->na        I   number of integer states
*            rtk->ns        O   number of valid satellite
*            rtk->tt        O   time difference between current and previous (s)
*            rtk->age       O   age of differential (s)
*            rtk->ambratio  O   ratio factor for ambiguity validation
*            rtk->x[]       IO  float states pre-filter and post-filter
*            rtk->P[]       IO  float covariance pre-filter and post-filter
*            rtk->xa[]      O   fixed states after AR
*            rtk->Pa[]      O   fixed covariance after AR
*            rtk->rr[0][]   IO  position rover ecef (m) (input for fixed mode,
*                               output of single solution)
*            rtk->rr[1][]   IO  position base ecef (m) (input for relative mode,
*                               output for moving-base mode)
*            rtk->az[0][s]  O   sat(s+1) azimuth angle rover (rad)
*            rtk->az[1][s]  O   sat(s+1) azimuth angle base  (rad)
*            rtk->el[0][s]  O   sat(s+1) elevation angle rover (rad)
*            rtk->el[1][s]  O   sat(s+1) elevation angle base  (rad)
*            rtk->vs[0][s]  O   sat(s+1) data valid single rover (0:invalid,1:valid)
*            rtk->vs[1][s]  O   sat(s+1) data valid single base  (0:invalid,1:valid)
*            rtk->sys  [s]    O  sat(s+1) system (SYS_???)
*            rtk->lams [s][f] O  sat(s+1) freq(f+1) wavelength (m)
*            rtk->resp [s][f] O  sat(s+1) freq(f+1) pseudorange residual (m)
*            rtk->resc [s][f] O  sat(s+1) freq(f+1) carrier-phase residual (m)
*            rtk->vsat [s][f] O  sat(s+1) freq(f+1) data vaild (0:invalid,1:valid)
*            rtk->fix  [s][f] O  sat(s+1) freq(f+1) ambiguity flag
*                                (0:nodata,1:float,2:fix,3:hold)
*            rtk->slip [s][f] O  sat(s+1) freq(f+1) slip flag
*                                (bit8-7:rcv1 LLI, bit6-5:rcv2 LLI,
*                                 bit2:parity unknown, bit1:slip)
*            rtk->lock [s][f] IO sat(s+1) freq(f+1) carrier lock count
*            rtk->outc [s][f] IO sat(s+1) freq(f+1) carrier outage count
*            rtk->slipc[s][f] IO sat(s+1) freq(f+1) cycle slip count
*            rtk->rejc [s][f] IO sat(s+1) freq(f+1) data reject count
*            rtk->gf[s]     IO  sat(s+1) geometry-free phase (L1-L2) (m)
*            rtk->nfix      IO  number of continuous fixes of ambiguity
*            rtk->neb       IO  bytes of error message buffer
*            rtk->errbuf    IO  error message buffer
*            rtk->tstr      O   time string for debug
*            rtk->opt       I   processing options
*          obsd_t *obs      I   observation data for an epoch
*                               obs[i].rcv=1:rover,2:reference
*                               sorted by receiver and satellte
*          int    nobs      I   number of observation data
*          nav_t  *nav      I   navigation messages
*          sbsstat_t *sbsstat I sbas status (NULL: no correction)
* return : solution status (0:error,1:fix,2:float,4:dgps,5:single)
* notes  : before calling function, base station position rtk->rr[1][] should
*          be properly set for relative mode except for moving-baseline
*-----------------------------------------------------------------------------*/
extern int rtkpos(rtk_t *rtk, const obsd_t *obs, int nobs, const nav_t *nav,
                  const sbsstat_t *sbsstat)
{
    prcopt_t *opt=&rtk->opt;
    double Qr[2][9]={{0}},Qv[2][9]={{0}},tt=0.0;
    int i,j,nu,nr,stat;
    char tstr[32];
    
    time2str(obs[0].time,tstr,2); strcpy(rtk->tstr,tstr+11);
    trace(3,"rtkpos : time=%s nobs=%d neph=%d\n",rtk->tstr,nobs,nav->n);
    trace(4,"obs=\n"); traceobs(4,obs,nobs);
    trace(5,"nav=\n"); tracenav(5,nav);
    
    /* set base staion position */
    if (opt->refpos<=2&&opt->mode!=PMODE_SINGLE&&opt->mode!=PMODE_MOVEB) {
        for (i=0;i<3;i++) rtk->rr[1][i]=opt->rb[i];
    }
    /* count rover/base station observations */
    for (nu=0;nu<nobs&&obs[nu].rcv==1;nu++) ;
    for (nr=0;nu+nr<nobs&&obs[nu+nr].rcv==2;nr++) ;
    
    /* estimate position/velocity of rover by single point positioning */
    if ((stat=rcvposv(rtk,obs,nu,nav,sbsstat,0,Qr[0],Qv[0]))==SOLQ_NONE) {
        if (!rtk->opt.dynamics) {
            outsolstat(rtk, SOLQ_NONE);
            return SOLQ_NONE;
        }
    }
    if (opt->mode==PMODE_SINGLE) {
        for (i=0;i<3;i++) {
            rtk->x[i]=rtk->rr[0][i];
            for (j=i;j<rtk->nx;j++) {
                rtk->P[i+j*rtk->nx]=rtk->P[j+i*rtk->nx]=j<3?Qr[0][i+j*3]:0.0;
            }
        }
        outsolstat(rtk,stat);
        return stat;
    }
    /* check number of data of base station and age of differential */
    if (nr==0) {
        errmsg(rtk,"no base station observation data\n");
        trace(2,"rtkpos : no base station observation data (%s)\n",rtk->tstr);
        return SOLQ_NONE;
    }
    rtk->age=timediff(obs[0].time,obs[nu].time);
    if (fabs(rtk->age)>opt->maxtdiff) {
        errmsg(rtk,"age of differential error (age=%.1f)\n",rtk->age);
        trace(1,"rtkpos : age of differential error (%s age=%.1f)\n",rtk->tstr,rtk->age);
        return SOLQ_NONE;
    }
    /* estimate position/velocity of base station for moving baseline */
    if (opt->mode==PMODE_MOVEB) {
        if (rcvposv(rtk,obs+nu,nr,nav,sbsstat,1,Qr[1],Qv[1])==SOLQ_NONE) {
            errmsg(rtk,"base station position error\n");
            trace(1,"rtkpos : base station position error (%s)\n",rtk->tstr);
            return SOLQ_NONE;
        }
        tt=timediff(rtk->time[0],rtk->time[1]);
        if (fabs(tt)>TTOL_MOVEB) {
            errmsg(rtk,"time sync error for moving-base (tt=%.2f)\n",tt);
            trace(2,"rtkpos : time sync error for moving-base (%s tt=%.2f)\n",rtk->tstr,tt);
            return SOLQ_NONE;
        }
    }
    /* relative positioning */
    stat=relpos(rtk,obs,nu,nr,nav);
    
    /* time-synchronized position of base station for moving baseline */
    if (opt->mode==PMODE_MOVEB) {
        for (i=0;i<3;i++) rtk->rr[1][i]+=rtk->rr[1][i+3]*tt;
    }
    outsolstat(rtk,stat);
    
    return stat;
}
